
#ifndef FIELD_ELEMENT_TEMPLATE_CLASS_DEFINED
#define FIELD_ELEMENT_TEMPLATE_CLASS_DEFINED

#include <cmath>
#include <iostream>
#include <cassert>
#include <vector>


using namespace std;

const int Zero_Polynomial[]={0};

class FiniteField
{
    public:
        unsigned int p;
        unsigned int deg;
        const int *poly;
        int number_of_field_elements;

        FiniteField(unsigned int prime, unsigned int degree, const int * polynom): p(prime), deg(degree), poly(polynom)
        {
            number_of_field_elements=pow(p,deg);
        };

        FiniteField(unsigned int prime): p(prime), deg(1), poly(Zero_Polynomial)
        {
            number_of_field_elements=pow(p,deg);
        };

        FiniteField(const FiniteField &f): p(f.p), deg(f.deg), poly(f.poly), number_of_field_elements(f.number_of_field_elements){};

     //   static const unsigned int degree = deg;
};
//  p: characteristic of field
//  deg: degree of field extension over prime field
//  poly: coefficients of monic polynomial defining the field extension; poly[i] is the coefficient of x^i; dynamic array of length deg

template <FiniteField* Fq_ptr>
class FieldElement;

template <FiniteField* Fq_ptr>
ostream & operator<< (ostream & os, const FieldElement<Fq_ptr> & a);

int mod(int a, unsigned int b)
{
    int r = a % (signed int) b;
    if (r<0)
        r += (signed int )b;
    return r;
}


template <FiniteField* Fq_ptr>
class FieldElement
{
    protected:
        int* coeffs;

    public:
        FieldElement(int* coefficients=(int*)0);
        FieldElement(const FieldElement & a);
        FieldElement(int a);


        FieldElement & operator=(const FieldElement & a);
        FieldElement & operator=(int a);

        FieldElement & operator+=(const FieldElement & a);
        FieldElement operator+(const FieldElement & a) const;
        FieldElement & operator-=(const FieldElement & a);
        FieldElement operator-(const FieldElement & a) const;
        FieldElement & operator*=(const FieldElement & FieldElement);
        FieldElement operator*(const FieldElement & FieldElement) const;
        FieldElement & operator/=(const FieldElement & FieldElement);
        FieldElement operator/(const FieldElement & FieldElement) const;

        bool is_zero(void) const;
        FieldElement inv(void) const;
        bool operator==(const FieldElement & a) const;
        bool operator==(int a) const;
        bool operator!=(const FieldElement & a) const;
        bool operator!=(int a) const;


        int & operator[](int i);

        ~FieldElement();

    friend ostream & operator<< <Fq_ptr>(ostream & os, const FieldElement & a);

};

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> pow(const FieldElement<Fq_ptr> & a, int e)
{
    FieldElement<Fq_ptr> y(1);

    if (e==0)
        return y;

    y = pow(a,e/2);
    if (e%2)
        return a*y*y;
    else
        return y*y;
}


template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr>::FieldElement(int* coefficients)
{
    coeffs = new int[Fq_ptr->deg];

    if (coefficients != (int*)0)
    {
        for (unsigned int i=0; i<Fq_ptr->deg; ++i)
            coeffs[i] = mod(coefficients[i], Fq_ptr->p);
    }
    else
    {
        for (unsigned int i=0; i<Fq_ptr->deg; ++i)
            coeffs[i] = 0;
    }
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr>::FieldElement(const FieldElement<Fq_ptr> & a)
{
    coeffs = new int[Fq_ptr->deg];

    for (unsigned int i=0; i<Fq_ptr->deg; ++i)
        coeffs[i] = a.coeffs[i];
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr>::FieldElement(int a)
{
    coeffs = new int[Fq_ptr->deg];

    coeffs[0] = mod(a, Fq_ptr->p);
    for (unsigned int i=1; i<Fq_ptr->deg; ++i)
            coeffs[i] = 0;

}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> & FieldElement<Fq_ptr>::operator=(const FieldElement<Fq_ptr> & a)
{
    if (this != &a)
    {
        for (unsigned int i=0; i<Fq_ptr->deg; ++i)
            coeffs[i] = a.coeffs[i];
    }

    return *this;
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> & FieldElement<Fq_ptr>::operator=(int a)
{
    coeffs[0] = mod(a, Fq_ptr->p);
    for (unsigned int i=1; i<Fq_ptr->deg; ++i)
            coeffs[i] = 0;

    return *this;
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> & FieldElement<Fq_ptr>::operator+=(const FieldElement<Fq_ptr> & a)
{
    for (unsigned int i=0; i<Fq_ptr->deg; ++i)
        this->coeffs[i] = mod(this->coeffs[i] + a.coeffs[i], Fq_ptr->p);

    return *this;
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> FieldElement<Fq_ptr>::operator+(const FieldElement<Fq_ptr> & a) const
{
    FieldElement<Fq_ptr> b(*this);
    return (b += a);
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> & FieldElement<Fq_ptr>::operator-=(const FieldElement<Fq_ptr> & a)
{
    for (unsigned int i=0; i<Fq_ptr->deg; ++i)
        this->coeffs[i] = mod(this->coeffs[i] - a.coeffs[i], Fq_ptr->p);

    return *this;
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> FieldElement<Fq_ptr>::operator-(const FieldElement<Fq_ptr> & a) const
{
    FieldElement<Fq_ptr> b(*this);
    return (b -= a);
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> & FieldElement<Fq_ptr>::operator*=(const FieldElement<Fq_ptr> & a)
{
    unsigned int deg = Fq_ptr->deg;
    unsigned int p = Fq_ptr->p;

    int* tmp = new int[2*deg]; // auxiliary polynomial
    for (unsigned int i=0; i<2*deg; ++i)
        tmp[i] = 0;

    for (unsigned int i=0; i<deg; ++i) // tmp = product of polynomials defining the elements
    {
        for (unsigned int j=0; j<deg; ++j)
            tmp[i+j] += coeffs[i] * a.coeffs[j];
    }

    for (unsigned int i=2*deg-1; i>=deg; --i) // tmp = tmp mod (poly, p)
    {
        tmp[i] = mod(tmp[i], p);
        if (tmp[i] != 0)
        {
            int lcf = tmp[i];
            tmp[i] = 0;
            for (unsigned int j=0; j<deg; ++j)
            {
                tmp[i-deg+j] = mod(tmp[i-deg+j] - lcf * Fq_ptr->poly[j], p);
            }
        }
    }


    for (unsigned int i=0; i<deg; ++i)
        this->coeffs[i] = mod(tmp[i], p);

    delete[] tmp;

    return *this;
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> FieldElement<Fq_ptr>::operator*(const FieldElement<Fq_ptr> & a) const
{
    FieldElement<Fq_ptr> b(*this);
    return (b *= a);
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> & FieldElement<Fq_ptr>::operator/=(const FieldElement<Fq_ptr> & a)
{
    return (*this *= a.inv());
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> FieldElement<Fq_ptr>::operator/(const FieldElement<Fq_ptr> & a) const
{
    FieldElement<Fq_ptr> b(*this);
    return (b /= a);
}

template <FiniteField* Fq_ptr>
bool FieldElement<Fq_ptr>::is_zero(void) const
{
    bool zero = true;
    for (int i=0; i<Fq_ptr->deg; ++i)
        zero = zero && (coeffs[i] == 0);

    return zero;
}


template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr> FieldElement<Fq_ptr>::inv(void) const
{
    unsigned int deg = Fq_ptr->deg;
    unsigned int p = Fq_ptr->p;

    assert(! this->is_zero()); // assert that element is invertible

    int* f = new int[deg+1]; // polynomial defining the field extension
    for (int i=0; i<deg; ++i)
        f[i] = Fq_ptr->poly[i];
    f[deg] = 1;

    int* a = new int[deg]; // polynomial defining element
    for (int i=0; i<deg; ++i)
        a[i] = coeffs[i];


    int* t = new int[deg]; // constant one polynomial
    t[0] = 1;
    for (int i=1; i<deg; ++i)
        t[i] = 0;

    int* v = new int[deg]; // constant zero polynomial
    for (int i=0; i<deg; ++i)
        v[i] = 0;

    int* q = new int[deg+1]; // auxiliary polynomials
    int* r = new int[deg+1];
    int* tmp = new int[2*deg+1];

    int degree_of_a = deg-1;
    while (a[degree_of_a]==0 && degree_of_a>=0)
        --degree_of_a;

    int degree_of_f = deg;

    while(degree_of_a >= 0)
    {
        int lcf_a = a[degree_of_a];

        // f = q*a + r
        for (int i=0; i<degree_of_f+1; ++i)
        {
            r[i] = f[i]; // r = f
            q[i] = 0;    // q = 0
        }
        for (int i=degree_of_f+1; i<deg+1; ++i)
        {
            r[i] = 0;
            q[i] = 0;
        }

        for (int i=degree_of_f; i>=degree_of_a; --i)
        {
            r[i] = mod(r[i], p);
            if (r[i] != 0)
            {
                int lcf_r = r[i];
                int factor = mod(lcf_r * (int)pow(lcf_a, p-2), p); // factor = lcf_r / lcf_a
                r[i] = 0;

                q[i-degree_of_a]= factor;

                for (int j=0; j<degree_of_a; ++j)
                {
                   r[i-degree_of_a+j] = mod(r[i-degree_of_a+j] - factor * a[j], p);
                }
            }
        }

        for (int i=0; i<=degree_of_a; ++i) // f = a
            f[i] = a[i];
        for (int i=degree_of_a+1; i<=degree_of_f; ++i)
            f[deg] = 0;
        degree_of_f = degree_of_a;

        for (int i=0; i<=degree_of_a; ++i) // a = r
            a[i] = r[i];
        while (degree_of_a>=0)
        {
            if (a[degree_of_a]==0)
                --degree_of_a;
            else
                break;
        }



        for (int i=0; i<2*deg+1; ++i) // tmp = 0
            tmp[i] = 0;

        for (int i=0; i<deg+1; ++i) // tmp = q * t
        {
            for (unsigned int j=0; j<deg; ++j)
                tmp[i+j] += q[i] * t[j];
        }

        for (int i=2*deg; i>=deg; --i) // tmp = tmp mod (poly, p)
        {
            tmp[i] = mod(tmp[i], p);
            if (tmp[i] != 0)
            {
                int lcf = tmp[i];
                tmp[i] = 0;
                for (unsigned int j=0; j<deg; ++j)
                {
                    tmp[i-deg+j] = mod(tmp[i-deg+j] - lcf * Fq_ptr->poly[j], p);
                }
            }
        }

        for (int i=0; i<deg; ++i) // tmp = v - tmp mod p
            tmp[i] = mod(v[i] - tmp[i], p);

        for (int i=0; i<deg; ++i) // v = t, t=tmp
        {
            v[i] = t[i];
            t[i] = tmp[i];
        }


    }

    // f[0] is the greatest common divisor of the polynomial defining the field extension and the polynomial defining the element
    int inv_of_gcd = mod((int)pow(f[0], p-2), p);

    for (int i=0; i<deg; ++i) // v = v / gcd
        v[i] = mod(v[i] * inv_of_gcd, p);

    FieldElement<Fq_ptr> inverse(v);

    delete[] f;
    delete[] a;
    delete[] t;
    delete[] v;
    delete[] q;
    delete[] r;
    delete[] tmp;

    return inverse;

}


template <FiniteField* Fq_ptr>
bool FieldElement<Fq_ptr>::operator==(const FieldElement & a) const
{
    bool eq = true;
    for (int i=0; i<Fq_ptr->deg; ++i)
        eq = eq && (coeffs[i] == a.coeffs[i]);

    return eq;
}

template <FiniteField* Fq_ptr>
bool FieldElement<Fq_ptr>::operator==(int a) const
{
    bool eq = true;
    for (int i=1; i<Fq_ptr->deg; ++i)
        eq = eq && (coeffs[i] == 0);

    eq = eq && coeffs[0]== mod(a, Fq_ptr->p);

    return eq;
}

template <FiniteField* Fq_ptr>
bool FieldElement<Fq_ptr>::operator!=(const FieldElement & a) const
{
    return !(*this == a);
}

template <FiniteField* Fq_ptr>
bool FieldElement<Fq_ptr>::operator!=(int a) const
{
    return !(*this == a);
}


template <FiniteField* Fq_ptr>
int & FieldElement<Fq_ptr>::operator[](int i)
{
    return coeffs[i];
}

template <FiniteField* Fq_ptr>
FieldElement<Fq_ptr>::~FieldElement()
{
    delete[] coeffs;
}



template <FiniteField* Fq_ptr>
ostream & operator<< (ostream & os, const FieldElement<Fq_ptr> & a)
{
    if (a==0)
        os << 0;


    else
    {
        bool before = false;
        if (a.coeffs[0] != 0)
        {
            os << a.coeffs[0];
            before = true;
        }

        if (Fq_ptr->deg >= 2 && a.coeffs[1] != 0)
        {
            if (before)
                os << " + ";
            if (a.coeffs[1] != 1)
                os << a.coeffs[1];
            os << " X";
            before = true;
        }

        for (unsigned int i=2; i<Fq_ptr->deg; ++i)
        {
            if (a.coeffs[i] != 0)
            {
                if (before)
                os << " + ";
                if (a.coeffs[i] != 1)
                    os << a.coeffs[i];
                os << " X^" << i;
                before = true;
            }
        }
    }
    return os;
}



#endif  // FIELD_ELEMENT_TEMPLATE_CLASS_DEFINED
